@@ -361,7 +361,6 @@ void clear_factors(int nfactors, mpz_t* pfactors[], int* pexponents[])
   Safefree(*pexponents);
 }
 
-
 void sigma(mpz_t res, mpz_t n, UV k)
 {
   mpz_t* factors;
@@ -645,65 +644,64 @@ void znprimroot(mpz_t root, mpz_t n)
   }
 }
 
+static const int32_t tau_table[] = {
+  0,1,-24,252,-1472,4830,-6048,-16744,84480,-113643,-115920,534612,-370944,-577738,401856,1217160,987136,-6905934,2727432,10661420,-7109760,-4219488,-12830688,18643272,21288960,-25499225,13865712,-73279080,24647168,128406630,-29211840,-52843168,-196706304,134722224,165742416,-80873520,167282496,-182213314,-255874080,-145589976,408038400,308120442,101267712,-17125708,-786948864,-548895690,-447438528
+};
+#define NTAU (sizeof(tau_table)/sizeof(tau_table[0]))
 void ramanujan_tau(mpz_t res, mpz_t n)
 {
-  mpz_t* factors;
-  mpz_t t, t1, t2, t3, t4;
-  int* exponents;
-  int i, nfactors;
+  mpz_t t, t1, t2, t3, t4, *factors;
+  int i, nfactors, *exponents;
   UV j, p2;
 
-  if (mpz_cmp_ui(n, 2) <= 0) {
-    if      (mpz_cmp_ui(n,2) == 0)  mpz_set_si(res, -24);
-    else if (mpz_cmp_ui(n,1) == 0)  mpz_set_si(res,   1);
-    else                            mpz_set_si(res,   0);
+  if (mpz_cmp_ui(n, NTAU) < 0) {
+    if (mpz_sgn(n) <= 0) mpz_set_si(res, 0);
+    else                 mpz_set_si(res, tau_table[mpz_get_ui(n)]);
     return;
   }
 
-  mpz_init(t);
-  mpz_init(t1); mpz_init(t2); mpz_init(t3); mpz_init(t4);
+  mpz_init(t); mpz_init(t1); mpz_init(t2); mpz_init(t3); mpz_init(t4);
   nfactors = factor(n, &factors, &exponents);
   for (i = 0; i < nfactors; i++) {
     /* t = tau(p) */
-    if (mpz_cmp_ui(factors[i], 2) == 0) {
-      mpz_set_si(t, -24);
+    if (mpz_cmp_ui(factors[i], NTAU) < 0) {
+      mpz_set_si(t, tau_table[mpz_get_ui(factors[i])]);
     } else {
       mpz_pow_ui(t, factors[i], 11);   mpz_add_ui(t, t, 1); /* sigma(t,f,11) */
       mpz_mul_ui(t1, t, 65);
       mpz_pow_ui(t, factors[i],  5);   mpz_add_ui(t, t, 1); /* sigma(t,f, 5) */
       mpz_mul_ui(t2, t, 691);
+      mpz_add(t1, t1, t2);
 
+      /* t1 in use. t2 accumulate.  t3, t4, t free. */
       mpz_sub_ui(t, factors[i], 1);
       mpz_tdiv_q_2exp(t, t, 1);
       p2 = mpz_get_ui(t);
-      mpz_set_ui(t3, 0);
+      mpz_set_ui(t2, 0);
       for (j = 1; j <= p2; j++) {
         mpz_set_ui(t, j);
-        sigma(t, t, 5);
-        mpz_mul_ui(t4, t, 2);
+        sigma(t3, t, 5);
         mpz_sub_ui(t, factors[i], j);
         sigma(t, t, 5);
-        mpz_mul(t4, t4, t);
-        mpz_add(t3, t3, t4);
+        mpz_mul(t4, t3, t);
+        mpz_add(t2, t2, t4);
       }
-      mpz_mul_ui(t3, t3, 691*252);
-      mpz_add(t, t1, t2);
-      mpz_sub(t, t, t3);
+      mpz_mul_ui(t2, t2, 2*691*252);
+      mpz_sub(t, t1, t2);
       mpz_tdiv_q_ui(t, t, 756);
     }
+    /* t holds tau(p), all other temps are free. */
 
     if (exponents[i] > 1) {
       mpz_pow_ui(t1, t, exponents[i]);
       if (exponents[i] == 2) {
         mpz_pow_ui(t2, factors[i], 11);
-        mpz_sub(t, t1, t2);
       } else if (exponents[i] == 3) {
         mpz_pow_ui(t2, factors[i], 11);
         mpz_mul(t2, t2, t);
         mpz_mul_ui(t2, t2, 2);
-        mpz_sub(t, t1, t2);
       } else {
-        /* t1 = t^e  t2 = sum,  t3 = prod,  t4 = temp */
+        /* t1 = t^e  t2 = neg sum,  t3 = prod,  t4 = temp */
         mpz_set_ui(t2, 0);
         for (j = 1; j <= (UV) (exponents[i]>>1); j++) {
           mpz_set_si(t3, (j&1) ? -1 : 1);
@@ -713,16 +711,15 @@ void ramanujan_tau(mpz_t res, mpz_t n)
           mpz_mul(t3, t3, t4);
           mpz_pow_ui(t4, t, exponents[i]-2*j);
           mpz_mul(t3, t3, t4);
-          mpz_add(t2, t2, t3);
+          mpz_sub(t2, t2, t3);
         }
-        mpz_add(t, t1, t2);
       }
+      mpz_sub(t, t1, t2);
     }
     mpz_set(factors[i], t);
   }
   mpz_product(factors, 0, nfactors-1);
   mpz_set(res, factors[0]);
   clear_factors(nfactors, &factors, &exponents);
-  mpz_clear(t1); mpz_clear(t2); mpz_clear(t3); mpz_init(t4);
-  mpz_clear(t);
+  mpz_clear(t1); mpz_clear(t2); mpz_clear(t3); mpz_init(t4); mpz_clear(t);
 }
