@@ -5,7 +5,7 @@ use Carp qw/croak confess carp/;
 
 BEGIN {
   $Math::Prime::Util::GMP::AUTHORITY = 'cpan:DANAJ';
-  $Math::Prime::Util::GMP::VERSION = '0.32';
+  $Math::Prime::Util::GMP::VERSION = '0.33';
 }
 
 # parent is cleaner, and in the Perl 5.10.1 / 5.12.0 core, but not earlier.
@@ -33,10 +33,12 @@ our @EXPORT_OK = qw(
                      is_mersenne_prime
                      is_llr_prime
                      is_proth_prime
+                     is_miller_prime
                      miller_rabin_random
                      lucas_sequence  lucasu  lucasv
                      primes
                      sieve_primes
+                     sieve_twin_primes
                      next_prime
                      prev_prime
                      trial_factor
@@ -182,7 +184,7 @@ Math::Prime::Util::GMP - Utilities related to prime numbers and factoring, using
 
 =head1 VERSION
 
-Version 0.32
+Version 0.33
 
 
 =head1 SYNOPSIS
@@ -313,12 +315,19 @@ tuning of this tradeoff, especially if performance is critical for
 by additional probable prime tests such as L</miller_rabin_random>
 and/or L</is_frobenius_underwood_pseudoprime>.
 
-As with L</is_prob_prime>, a BPSW test is first performed.  If this
-indicates "probably prime" then a small number of Miller-Rabin tests
-with random bases are performed.  For numbers under 200 bits, a quick
+As with L</is_prob_prime>, a BPSW test is first performed.  This is
+deterministic for all 64-bit numbers.  Next, if the number is a
+Proth or LLR form, then a proof is constructed.  If the result is
+still "probably prime" and the input is smaller than the
+Sorenson/Webster (2015) deterministic Miller-Rabin limit
+(approximately 82 bits) then the 11 or 12 Miller-Rabin tests are
+performed and the result is confirmed.  For larger inputs that are
+still "probably prime" but under 200 bits, a quick
 BLS75 C<n-1> primality proof is attempted.  This is tuned to give up
 if the result cannot be quickly determined, and results in success
 rates of ~80% at 80 bits, ~30% at 128 bits, and ~13% at 160 bits.
+Lastly, for results still "probably prime", a small number of
+Miller-Rabin tests with random bases are performed.
 
 The result is that many numbers will return 2 (definitely prime),
 and the numbers that return 1 (probably prime) have gone through
@@ -343,10 +352,12 @@ Takes a positive number as input and returns back either 0 (composite),
 2 (definitely prime), or 1 (probably prime).  A great deal of effort is
 taken to return either 0 or 2 for all numbers.
 
-The current method first uses BPSW and a small number of Miller-Rabin
-tests with random bases to weed out composites and provide a
-deterministic answer for tiny numbers (under C<2^64>).  A quick BLS75
-C<n-1> test is attempted, followed by ECPP.
+The current method first uses BPSW to find composites and provide a
+deterministic answer for tiny numbers (under C<2^64>).  If no
+certificate is required, LLR and Proth tests can be run, and small
+numbers (under approximately C<2^82>) can be satisfied with a
+deterministic Miller-Rabin test.  If the result is still not determined,
+a quick  BLS75 C<n-1> test is attempted, followed by ECPP.
 
 The time required for primes of different input sizes on a circa-2009
 workstation averages about C<3ms> for 30-digits, C<5ms> for 40-digit,
@@ -634,6 +645,25 @@ While not as fast as the Lucas-Lehmer test for Mersenne
 numbers, it is almost as fast as a single strong pseudoprime test (i.e.
 Miller-Rabin test) while giving a certain answer.
 
+=head2 is_miller_prime
+
+  say "$n is definitely prime" if is_miller_prime($n);
+  say "$n is definitely prime assuming the GRH" if is_miller_prime($n, 1);
+
+Takes a positive number as input, and returns 1 if the input passes the
+deterministic Miller test.  An optional second argument indicates whether
+the Generalized Riemann Hypothesis should be assumed, and defaults to 0.
+Setting the verbose flag to 2 or higher will show how many bases are used.
+The unconditional test is exponential time, while the conditional test
+(assuming the GRH) is polynomial time.
+
+This is a very slow method in practice, and generally should not be used.
+The asymptotic complexity of the GRH version is good in theory, matching
+ECPP, but in practice it is much slower.  The number of bases used by
+the unconditional test grows quite rapidly, impractically many past about
+160 bits, and overflows a 64-bit integer at 456 bits -- sizes that are
+trivial for the unconditional APR-CL and ECPP tests.
+
 =head2 is_nminus1_prime
 
   say "$n is definitely prime" if is_nminus1_prime($n);
@@ -703,6 +733,18 @@ that want to apply their own primality or other tests, and wish to have a
 list of values in the range with no small factors.  This is quite common
 for applications involving prime gaps.
 
+=head2 sieve_twin_primes
+
+  my @primes = sieve_twin_primes(2**1000, 2**1000 + 500000);
+
+Given two arguments C<low> and C<high>, this returns each lower twin prime
+in the interval (inclusive).  The result is a list, not a reference.
+
+This does a partial sieve of the range, removes any non-twin candidates,
+then checks that each pair are both BPSW probable primes.  This is
+substantially more efficient than sieving for all primes followed by
+removing those that are not twin primes.
+
 
 =head2 next_prime
 
